import numpy as np
from numba import njit, prange

@njit
def track_reconstruction(hits: np.ndarray):
    n_layers = hits.shape[0]
    A = np.zeros((n_layers, 2), dtype=np.float32)
    B = np.zeros((n_layers, 2), dtype=np.float32)

    for i in range(n_layers):
        A[i, 0] = hits[i, 2]
        A[i, 1] = 1
        B[i, 0] = hits[i, 0]
        B[i, 1] = hits[i, 1]

    AtA_inv = np.linalg.inv(A.T @ A)
    AtB = A.T @ B
    params = AtA_inv @ AtB

    slope_x = params[0, 0]
    intercept_x = params[1, 0]
    slope_y = params[0, 1]
    intercept_y = params[1, 1]

    theta_x = np.degrees(np.arctan(slope_x))
    theta_y = np.degrees(np.arctan(slope_y))

    return np.array([intercept_x, intercept_y, theta_x, theta_y], dtype=np.float32)


import numpy as np
import matplotlib.pyplot as plt
from numba import njit, prange
from coordinates import det2earth, earth2det, cart2slope
from tracking import track_reconstruction
from flux import effective_area, solid_angle
import time


@njit
def homogenous_generator(theta_max_deg: float, zenith_boresight_deg: float, azimuth_boresight_deg: float, half_length_x: float, half_length_y: float):
    theta_max_rad = np.radians(theta_max_deg)
    cos_theta = np.random.uniform(np.cos(theta_max_rad), 1)
    theta = np.arccos(cos_theta)
    phi = np.random.uniform(0, 2 * np.pi)

    x_dir_det = np.sin(theta) * np.cos(phi)
    y_dir_det = np.sin(theta) * np.sin(phi)
    z_dir_det = np.cos(theta)

    x_pos_det = 0
    y_pos_det = 0
    z_pos_det = 0

    x_pp = np.random.uniform(-half_length_x, half_length_x)
    y_pp = np.random.uniform(-half_length_y, half_length_y)
    z_pp = 0

    x_pos_det += -x_pp * np.sin(phi) - y_pp * np.cos(theta) * np.cos(phi) + z_pp * np.sin(theta) * np.cos(phi)
    y_pos_det += x_pp * np.cos(phi) - y_pp * np.cos(theta) * np.sin(phi) + z_pp * np.sin(theta) * np.sin(phi)
    z_pos_det += y_pp * np.sin(theta) + z_pp * np.cos(theta)

    x_dir_earth, y_dir_earth, z_dir_earth = det2earth(x_dir_det, y_dir_det, z_dir_det, zenith_boresight_deg, azimuth_boresight_deg)
    x_pos_earth, y_pos_earth, z_pos_earth = det2earth(x_pos_det, y_pos_det, z_pos_det, zenith_boresight_deg, azimuth_boresight_deg)

    return np.array([x_pos_earth, y_pos_earth, z_pos_earth]), np.array([x_dir_earth, y_dir_earth, z_dir_earth])


@njit
def detection_simulation(layer_z: np.ndarray, pixel_length_x: float, pixel_length_y: float, layer_half_length_x: np.ndarray, layer_half_length_y: np.ndarray, zenith_boresight_deg: float, azimuth_boresight_deg: float, particle_pos_earth: np.ndarray, particle_dir_earth: np.ndarray, mode: int):

    particle_dir_det = earth2det(particle_dir_earth[0], particle_dir_earth[1], particle_dir_earth[2], zenith_boresight_deg, azimuth_boresight_deg)
    particle_pos_det = earth2det(particle_pos_earth[0], particle_pos_earth[1], particle_pos_earth[2], zenith_boresight_deg, azimuth_boresight_deg)

    theta_x = np.arctan2(particle_dir_det[0], particle_dir_det[2])
    theta_y = np.arctan2(particle_dir_det[1], particle_dir_det[2])

    x = np.tan(theta_x) * (layer_z - particle_pos_det[2]) + particle_pos_det[0]
    y = np.tan(theta_y) * (layer_z - particle_pos_det[2]) + particle_pos_det[1]

    if mode == 0:
        x = np.round(x / pixel_length_x) * pixel_length_x
        y = np.round(y / pixel_length_y) * pixel_length_y
    elif mode == 1:
        x = np.floor(x / pixel_length_x) * pixel_length_x + pixel_length_x / 2
        y = np.floor(y / pixel_length_y) * pixel_length_y + pixel_length_y / 2
    else:
        raise ValueError("mode must be 0 or 1")

    hits = np.zeros((len(layer_z), 3), dtype=np.float32)
    hit = True
    for i in range(len(layer_z)):
        if np.abs(x[i]) <= layer_half_length_x[i] and np.abs(y[i]) <= layer_half_length_y[i]:
            hits[i, 0] = x[i]
            hits[i, 1] = y[i]
            hits[i, 2] = layer_z[i]
        else:
            hits[i, 0] = np.nan
            hits[i, 1] = np.nan
            hits[i, 2] = layer_z[i]
            hit = False

    return hits, hit


@njit(parallel=True)
def run_homogenous_simulation(num_particles: int, theta_max_deg: float, zenith_boresight_deg: float, azimuth_boresight_deg: float, half_length_x: float, half_length_y: float, layer_z: np.ndarray, pixel_length_x: float, pixel_length_y: float, layer_half_length_x: np.ndarray, layer_half_length_y: np.ndarray, mask=None):
    all_hit = np.zeros(num_particles, dtype=np.bool_)
    all_measured = np.empty((num_particles, 4), dtype=np.float32)
    all_simulated = np.empty((num_particles, 3), dtype=np.float32)
    for i in prange(num_particles):
        particle_pos_earth, particle_dir_earth = homogenous_generator(theta_max_deg, zenith_boresight_deg, azimuth_boresight_deg, half_length_x, half_length_y)
        hits, hit = detection_simulation(layer_z, pixel_length_x, pixel_length_y, layer_half_length_x, layer_half_length_y, zenith_boresight_deg, azimuth_boresight_deg, particle_pos_earth, particle_dir_earth, mode=1)
        if mask is not None:
            hits = hits[mask]
        all_hit[i] = hit
        all_measured[i] = track_reconstruction(hits) if hit else np.array([np.nan, np.nan, np.nan, np.nan], dtype=np.float32)
        r, theta_x, theta_y = cart2slope(particle_dir_earth[0], particle_dir_earth[1], particle_dir_earth[2])
        all_simulated[i] = np.array([r, np.degrees(theta_x), np.degrees(theta_y)], dtype=np.float32)

    return all_simulated[all_hit], all_measured[all_hit]


@njit(parallel=True)
def compute_basis(simulated_hits: np.ndarray, detected_hits: np.ndarray, opening_angle_deg=25, basis=None):
    detected_hits = np.round(detected_hits, decimals=1)
    theta_x_unique = np.unique(detected_hits[:, 2])
    theta_y_unique = np.unique(detected_hits[:, 3])

    angle = opening_angle_deg
    mrad = int(np.radians(angle) * 1000)
    nx = np.arange(- mrad, mrad + 1)
    ny = np.arange(- mrad, mrad + 1)

    cx = np.amin(nx)
    cy = np.amin(ny)

    simulated = simulated_hits[:, 1:3]
    detected = detected_hits[:, 2:4]

    if basis is None:
        basis = np.zeros((len(theta_x_unique), len(theta_y_unique), len(nx), len(ny)), dtype=np.int32)

    for i in prange(simulated.shape[0]):
        dx = np.zeros(8, dtype=np.float32)
        dy = np.zeros(8, dtype=np.float32)

        ix = np.zeros(8, dtype=np.int32)
        iy = np.zeros(8, dtype=np.int32)

        sx = np.zeros(8, dtype=np.float32)
        sy = np.zeros(8, dtype=np.float32)

        dx_ = detected[i, 0]
        dy_ = detected[i, 1]

        dx[0], dy[0] = dx_, dy_
        dx[1], dy[1] = -dx_, dy_
        dx[2], dy[2] = dx_, -dy_
        dx[3], dy[3] = -dx_, -dy_
        dx[4], dy[4] = dy_, dx_
        dx[5], dy[5] = -dy_, dx_
        dx[6], dy[6] = dy_, -dx_
        dx[7], dy[7] = -dy_, -dx_

        for j in range(8):
            ix[j] = np.searchsorted(theta_x_unique, dx[j])
            iy[j] = np.searchsorted(theta_y_unique, dy[j])

        x_ = np.round(np.radians(simulated[i, 0]) * 1000)
        y_ = np.round(np.radians(simulated[i, 1]) * 1000)

        sx[0], sy[0] = x_, y_
        sx[1], sy[1] = -x_, y_
        sx[2], sy[2] = x_, -y_
        sx[3], sy[3] = -x_, -y_
        sx[4], sy[4] = y_, x_
        sx[5], sy[5] = -y_, x_
        sx[6], sy[6] = y_, -x_
        sx[7], sy[7] = -y_, -x_

        sx -= cx
        sy -= cy

        for j in range(8):
            basis[ix[j], iy[j], int(sx[j]), int(sy[j])] += 1


    return basis, theta_x_unique, theta_y_unique, nx, ny


if __name__ == "__main__":
    layer_z = np.array([-750, -250, 250, 750], dtype=np.float32)
    pixel_length_x = 50
    pixel_length_y = 50
    layer_half_length_x = np.array([300, 200, 200, 300], dtype=np.float32)
    layer_half_length_y = np.array([300, 200, 200, 300], dtype=np.float32)
    zenith_boresight_deg = 0.0
    azimuth_boresight_deg = 0.0
    theta_max_deg = 30.0
    half_length_x = 300
    half_length_y = 300

    flux = 1
    solid_angle_ = 2 * np.pi * (1 - np.cos(np.radians(theta_max_deg)))
    area = (2 * half_length_x / 1000) * (2 * half_length_y / 1000)
    exposure_time = 3650 * 24 * 3600  # 3650 days in seconds
    mask = np.array([True, False, True, True])
    geometric_factor = solid_angle_ * area

    num_particles = int(flux * solid_angle_ * area * exposure_time)

    angle = 25
    mrad = int(np.radians(angle) * 1000)
    theta_x_mrad = np.arange(-mrad, mrad + 1)
    theta_y_mrad = np.arange(-mrad, mrad + 1)

    theta_x_rad, theta_y_rad = np.meshgrid(theta_x_mrad / 1000, theta_y_mrad / 1000)

    area = np.zeros(theta_x_rad.shape, dtype=np.float32)
    sa = np.zeros(theta_x_rad.shape, dtype=np.float32)
    for i in range(theta_x_rad.shape[0]):
        for j in range(theta_x_rad.shape[1]):
            area[i, j] = effective_area(theta_x_rad[i, j], theta_y_rad[i, j], layer_z, layer_half_length_x, layer_half_length_y) / 1000 / 1000
            sa[i, j] = solid_angle(theta_x_rad[i, j], theta_y_rad[i, j], 1 / 1000, 1 / 1000)

    geometric_factor_array = area * sa


    runs = 1
    while True:
        print(f"Run {runs}")
        total_counts = 0
        target_counts = 100_000_000
        simulated_hits, detected_hits = run_homogenous_simulation(num_particles, theta_max_deg, zenith_boresight_deg, azimuth_boresight_deg, half_length_x, half_length_y, layer_z, pixel_length_x, pixel_length_y, layer_half_length_x, layer_half_length_y, mask=mask)
        basis, theta_x_unique, theta_y_unique, nx, ny = compute_basis(simulated_hits, detected_hits)

        total_counts += simulated_hits.shape[0]
        total_time = 0
        i = 0
        while True:
            start_time = time.time()
            simulated_hits, detected_hits = run_homogenous_simulation(num_particles, theta_max_deg, zenith_boresight_deg, azimuth_boresight_deg, half_length_x, half_length_y, layer_z, pixel_length_x, pixel_length_y, layer_half_length_x, layer_half_length_y, mask=mask)
            basis, theta_x_unique, theta_y_unique, nx, ny = compute_basis(simulated_hits, detected_hits, basis=basis)
            total_counts += simulated_hits.shape[0]
            end_time = time.time()
            total_time += end_time - start_time
            print(f"Iteration {i+1}: Processed {simulated_hits.shape[0]} hits in {end_time - start_time:.2f} seconds. Total counts: {total_counts}. Total time: {total_time:.2f} seconds. Average time per iteration: {total_time / (i + 1):.2f} seconds.")
            i += 1
            if total_counts >= target_counts:
                break

        # if exists, load and add
        try:
            old_basis = np.load("monte_carlo_basis_2.npy")
            basis += old_basis
            print("Loaded existing basis and added to current basis.")
        except FileNotFoundError:
            print("No existing basis found. Saving current basis.")

        np.save("monte_carlo_basis_2.npy", basis)

        image = np.sum(basis, axis=(0, 1))

        particle_counts = np.sum(basis)
        image = image.astype(np.float64)
        image[geometric_factor_array > 0] /= geometric_factor_array[geometric_factor_array > 0]

        mean = np.mean(image[image > 0])
        std = np.std(image[image > 0])

        with open("monte_carlo_info_2.txt", "a") as f:
            string_ = f"{std/mean*100}, {particle_counts}\n"
            f.write(string_)

        print(f"{std/mean * 100:.2f}% uncertainty with {particle_counts:.2E} particles")

        measured_angles_x, measured_angles_y = np.meshgrid(theta_x_unique, theta_y_unique)
        measured_angles = np.array([measured_angles_x, measured_angles_y])
        measured_angles = np.radians(measured_angles) * 1000

        incident_angles_x, incident_angles_y = np.meshgrid(nx, ny)
        incident_angles = np.array([incident_angles_x, incident_angles_y])

        dictionary = {
            "basis": basis,
            "measured_angles": measured_angles,
            "incident_angles": incident_angles,
            "geometric_factor": geometric_factor_array
        }

        np.savez_compressed("9449_2.npz", **dictionary)
        print("Saved 9449_2.npz")

        runs += 1

        if np.amax(basis) >= 2_000_000_000:
            break
    # simulated_hits, detected_hits = run_simulation(num_particles)
    # np.save("../../simulation_1flux_36500days2.npy", detected_hits)




import numpy as np
from numba import njit


@njit
def cart2slope(x: float, y: float, z: float):
    r = np.sqrt(x**2 + y**2 + z**2)
    theta_x = np.arctan2(x, z)
    theta_y = np.arctan2(y, z)

    return r, theta_x, theta_y


@njit
def cart2slope_vec(p: np.ndarray):
    # Reshape to (-1, 3) to handle arbitrary dimensions
    original_shape = p.shape[:-1]
    x_reshaped = p.reshape(-1, 3)

    x0 = x_reshaped[:, 0]
    x1 = x_reshaped[:, 1]
    x2 = x_reshaped[:, 2]

    r = np.sqrt(x0 ** 2 + x1 ** 2 + x2 ** 2)
    theta_x = np.degrees(np.arctan2(x0, x2))
    theta_y = np.degrees(np.arctan2(x1, x2))

    v = np.empty((x_reshaped.shape[0], 3))
    v[:, 0] = r
    v[:, 1] = theta_x
    v[:, 2] = theta_y

    # Reshape back to the original dimensions
    return v.reshape(*original_shape, 3)


@njit
def slope2cart(r: float, theta_x_deg: float, theta_y_deg: float):
    theta_x = np.radians(theta_x_deg)
    theta_y = np.radians(theta_y_deg)

    x = r * np.tan(theta_x) / np.sqrt(1 + np.tan(theta_x) ** 2 + np.tan(theta_y) ** 2)
    y = r * np.tan(theta_y) / np.sqrt(1 + np.tan(theta_x) ** 2 + np.tan(theta_y) ** 2)
    z = r / np.sqrt(1 + np.tan(theta_x) ** 2 + np.tan(theta_y) ** 2)

    return x, y, z


@njit
def slope2cart_vec(p: np.ndarray):
    # Reshape to (-1, 3) to handle arbitrary dimensions
    original_shape = p.shape[:-1]
    p_reshaped = p.reshape(-1, 3)

    r = p_reshaped[:, 0]
    theta_x = p_reshaped[:, 1]
    theta_y = p_reshaped[:, 2]

    x = r * np.tan(theta_x) / np.sqrt(1 + np.tan(theta_x) ** 2 + np.tan(theta_y) ** 2)
    y = r * np.tan(theta_y) / np.sqrt(1 + np.tan(theta_x) ** 2 + np.tan(theta_y) ** 2)
    z = r / np.sqrt(1 + np.tan(theta_x) ** 2 + np.tan(theta_y) ** 2)

    cartesian = np.empty((p_reshaped.shape[0], 3))
    cartesian[:, 0] = x
    cartesian[:, 1] = y
    cartesian[:, 2] = z

    # Reshape back to the original dimensions
    return cartesian.reshape(*original_shape, 3)


@njit
def rotational_matrix(zenith_det_deg: float, azimuth_det_deg: float):
    zenith = np.radians(zenith_det_deg)
    azimuth = -np.radians(azimuth_det_deg)

    R = np.array([
        [np.cos(azimuth), -np.sin(azimuth), 0],
        [np.sin(azimuth) * np.cos(zenith), np.cos(azimuth) * np.cos(zenith), -np.sin(zenith)],
        [np.sin(azimuth) * np.sin(zenith), np.sin(zenith) * np.cos(azimuth), np.cos(zenith)]
    ])

    return R


@njit
def earth2det(x: float, y: float, z: float, zenith_det_deg: float, azimuth_det_deg: float):
    X = np.array([x, y, z])
    R = rotational_matrix(zenith_det_deg, azimuth_det_deg)

    Xp = R @ X
    xp, yp, zp = Xp[0], Xp[1], Xp[2]

    return xp, yp, zp


@njit
def earth2det_vec(p: np.ndarray, zenith_det_deg: float, azimuth_det_deg: float):
    # Reshape to (-1, 3) to handle arbitrary dimensions
    original_shape = p.shape[:-1]
    p_reshaped = p.reshape(-1, 3)

    R = rotational_matrix(zenith_det_deg, azimuth_det_deg)

    Xp = R @ p_reshaped.T
    xp = Xp[0, :]
    yp = Xp[1, :]
    zp = Xp[2, :]

    dct = np.empty((p_reshaped.shape[0], 3))
    dct[:, 0] = xp
    dct[:, 1] = yp
    dct[:, 2] = zp

    # Reshape back to the original dimensions
    return dct.reshape(*original_shape, 3)


@njit
def det2earth(xp: float, yp: float, zp: float, zenith_det_deg: float, azimuth_det_deg: float):
    Xp = np.array([xp, yp, zp])
    R = rotational_matrix(zenith_det_deg, azimuth_det_deg)
    Rt = R.T

    X = Rt @ Xp
    x, y, z = X[0], X[1], X[2]

    return x, y, z


@njit
def det2earth_vec(p: np.ndarray, zenith_det_deg: float, azimuth_det_deg: float):
    # Reshape to (-1, 3) to handle arbitrary dimensions
    original_shape = p.shape[:-1]
    p_reshaped = p.reshape(-1, 3)

    R = rotational_matrix(zenith_det_deg, azimuth_det_deg)
    Rt = R.T

    X = Rt @ p_reshaped.T
    x = X[0, :]
    y = X[1, :]
    z = X[2, :]

    earth = np.empty((p_reshaped.shape[0], 3))
    earth[:, 0] = x
    earth[:, 1] = y
    earth[:, 2] = z

    # Reshape back to the original dimensions
    return earth.reshape(*original_shape, 3)


@njit
def zenith_azimuth_deg(theta_x: float, theta_y: float, zenith_det_deg: float, azimuth_det_deg: float):
    zenith = np.radians(zenith_det_deg)

    xp, yp, zp = slope2cart(1.0, theta_x, theta_y)

    x, y, z = det2earth(xp, yp, zp, zenith_det_deg, azimuth_det_deg)

    cos_zenith = -np.sin(zenith) * yp + np.cos(zenith) * zp
    zenith_deg = np.degrees(np.arccos(cos_zenith))

    azimuth_deg = np.degrees(np.arctan2(-x, y))

    if azimuth_deg < 0:
        azimuth_deg += 360.0

    return zenith_deg, azimuth_deg


@njit
def zenith_azimuth_deg_vec(p: np.ndarray, zenith_det_deg: float, azimuth_det_deg: float):
    # Reshape to (-1, 2) to handle arbitrary dimensions
    original_shape = p.shape[:-1]
    p_reshaped = p.reshape(-1, 2)

    zenith = np.radians(zenith_det_deg)

    theta_x = p_reshaped[:, 0]
    theta_y = p_reshaped[:, 1]

    xp, yp, zp = slope2cart_vec(np.column_stack((np.ones(theta_x.shape), theta_x, theta_y))).T

    x, y, z = det2earth_vec(np.column_stack((xp, yp, zp)), zenith_det_deg, azimuth_det_deg).T

    cos_zenith = -np.sin(zenith) * yp + np.cos(zenith) * zp
    zenith_deg = np.degrees(np.arccos(cos_zenith))

    azimuth_deg = np.degrees(np.arctan2(-x, y))

    for i in range(azimuth_deg.shape[0]):
        if azimuth_deg[i] < 0:
            azimuth_deg[i] += 360.0

    za = np.empty((p_reshaped.shape[0], 2))
    za[:, 0] = zenith_deg
    za[:, 1] = azimuth_deg

    # Reshape back to the original dimensions
    return za.reshape(*original_shape, 2)


if __name__ == "__main__":
    theta_x = 0
    theta_y = 0

    xp, yp, zp = slope2cart(1.0, theta_x, theta_y)

    print(f"xp: {xp}, yp: {yp}, zp: {zp}")

    zenith_det_deg = 10
    azimuth_det_deg = 0

    zenith_deg, azimuth_deg = zenith_azimuth_deg(theta_x, theta_y, zenith_det_deg, azimuth_det_deg)

    print(f"zenith_deg: {zenith_deg}, azimuth_deg: {azimuth_deg}")


import numpy as np
import matplotlib.pyplot as plt
from numba import njit


@njit
def array2combo(array: np.ndarray, layer_z: np.ndarray):

    counts = np.zeros(len(layer_z), dtype=np.int32)

    for i in range(len(layer_z)):
        counts[i] = np.sum(array[:, 1] == layer_z[i])

    combos = 1
    for c in counts:
        if c > 1:
            combos *= c

    L = len(layer_z)
    result = np.empty((combos, L, 2), dtype=array.dtype)
    index = np.zeros(L, dtype=np.int32)

    for i in range(combos):
        for j in range(L):
            layer_hits = array[array[:, 1] == layer_z[j]]
            if len(layer_hits) > 0:
                result[i, j] = layer_hits[index[j] % len(layer_hits)]
            else:
                result[i, j] = np.array([np.nan, layer_z[j]])

        for j in range(L - 1, -1, -1):
            if counts[j] > 0:
                index[j] += 1
                if index[j] < counts[j]:
                    break
                index[j] = 0

    return result


@njit
def intercept_(pos1: np.ndarray, pos2: np.ndarray):
    a = np.arctan2(pos1[0] - pos2[0], pos1[1] - pos2[1])
    b = pos1[0] - np.tan(a) * pos1[1]

    return np.array([a, b])


@njit
def multiple_intercept_(positions: np.ndarray, delta_b: float):
    n = positions.shape[0]
    intercepts_ = np.zeros((n, n, 2), dtype=float)
    z_ = np.zeros((n, n), dtype=float)

    for i in range(n):
        for j in range(n):
            if i <= j:
                intercepts_[i, j, :] = (np.nan, np.nan)
                z_[i, j] = np.nan
            else:
                if np.isnan(positions[i, 0]) or np.isnan(positions[j, 0]):
                    intercepts_[i, j, :] = (np.nan, np.nan)
                    z_[i, j] = np.nan
                else:
                    intercepts_[i, j, :] = intercept_(positions[i], positions[j])
                    z_[i, j] = np.abs(positions[j, 1] - positions[i, 1])

    intercepts_ = intercepts_.reshape(-1, 2)
    z_ = z_.reshape(-1)

    mask = np.ones(intercepts_.shape[0], dtype=np.bool_)
    for i in range(intercepts_.shape[0]):
        if np.isnan(intercepts_[i, 0]) or np.isnan(intercepts_[i, 1]):
            mask[i] = False

    intercepts_ = intercepts_[mask]
    z_ = z_[mask]

    centroid_a = np.average(intercepts_[:, 0], weights=1/z_)
    centroid_b = np.mean(intercepts_[:, 1])

    centroid = np.array([centroid_a, centroid_b])

    intercepts_ -= centroid

    b_lsq = np.sum((intercepts_[:, 1]/delta_b)**2)
    a_lsq = np.sum((intercepts_[:, 0]/(delta_b/z_))**2)

    lsq = a_lsq + b_lsq
    return lsq






import numpy as np
from numba import njit
from tqdm import tqdm
import os


@njit
def pcnt_correction(pcnt):
    timestamp = np.zeros_like(pcnt)
    previous_timestamp = None

    for n, pcnt in enumerate(pcnt):
        if previous_timestamp is None:
            previous_timestamp = pcnt
            timestamp[n] = pcnt
        else:
            if pcnt < previous_timestamp:
                if previous_timestamp - pcnt > 2**17:
                    timestamp[n] = pcnt + 2**20
                else:
                    timestamp[n] = previous_timestamp
            else:
                timestamp[n] = pcnt

            previous_timestamp = timestamp[n]

    return timestamp


def get_timestamp(pcnt, tcnt):
    pcnt = pcnt.astype(np.uint64)
    pcnt = pcnt << 32
    tcnt = tcnt.astype(np.uint64)

    timestamp = pcnt + tcnt
    return timestamp


def get_layer(boardID, mapping):
    layerID = np.zeros(len(boardID))
    for N in range(len(boardID)):
        ID = boardID[N]
        layerID[N] = mapping[int(ID)-1]

    return layerID


def get_unique_id(boardID, channelID, mapping):

    unique_id = np.zeros_like(boardID, dtype=np.uint16)
    for n in range(len(boardID)):
        unique_id[n] = mapping[int(boardID[n]-1), int(channelID[n])]

    return unique_id


# 0 - Type
# 1 - Board ID
# 2 - Channel ID
# 3 - EVT
# 4 - Timestamp
# 5 - PCNT
# 6 - TCNT
# 7 - PWIDTH


@njit
def label(timestamp, event_threshold=75, hit_threshold=15):
    current_label = 0
    initial_timestamp = timestamp[0]
    previous_timestamp = timestamp[0]
    labels = np.zeros(len(timestamp), dtype=np.uint64)


    for n in range(1, len(timestamp)):
        current_timestamp = timestamp[n]

        if current_timestamp - previous_timestamp > hit_threshold or current_timestamp - initial_timestamp > event_threshold:
            current_label += 1
            initial_timestamp = current_timestamp

        previous_timestamp = current_timestamp
        labels[n] = current_label

    return labels


def event_selection(event):
    layer_ids = event["LAYERID"]
    layer_id, counts = np.unique(layer_ids, return_counts=True)

    if len(layer_id) < 4:
        return False
    elif counts[0] > 3 or counts[1] > 3 or counts[2] > 3 or counts[3] > 2 or len(layer_ids) > 8:
        return False
    else:
        return True



def event_grouping(data, event_selection):
    # Create a structured array for all hits
    dtype = [
        ("EVENTID", data["EVENTID"].dtype),
        ("BOARDID", data["BOARDID"].dtype),
        ("CHANNELID", data["CHANNELID"].dtype),
        ("LAYERID", data["LAYERID"].dtype),
        ("UNIQUEID", data["UNIQUEID"].dtype),
        ("TIMESTAMP", data["TIMESTAMP"].dtype),
        ("PCNT", data["PCNT"].dtype),
        ("TCNT", data["TCNT"].dtype),
        ("PWIDTH", data["PWIDTH"].dtype),
    ]
    hits = np.empty(len(data["EVENTID"]), dtype=dtype)
    for key in hits.dtype.names:
        hits[key] = data[key]

    # Group by EVENTID and filter events
    event_ids, start_idx, counts = np.unique(hits["EVENTID"], return_index=True, return_counts=True)
    keep_mask = np.zeros(len(hits), dtype=bool)

    for eid, idx, count in zip(event_ids, start_idx, counts):
        event_slice = slice(idx, idx + count)
        event = {k: hits[k][event_slice] for k in hits.dtype.names if k != "EVENTID"}
        if event_selection(event):
            keep_mask[event_slice] = True

    return hits[keep_mask]


def parser(dir, file, unique_id_mapping, layer_mapping, event_selection, event_threshold=75, hit_threshold=75):
    filepath = os.path.join(dir, file)

    # First pass: count valid lines (skip those with "Frame")
    valid_line_count = 0
    total_line_count = 0
    with open(filepath, "r") as f:
        for line in f:
            if not line.startswith("#Frame"):
                valid_line_count += 1
            total_line_count += 1


    # Preallocate arrays with the exact number of rows needed
    BOARDID   = np.empty(valid_line_count, dtype=np.uint8)
    CHANNELID = np.empty(valid_line_count, dtype=np.uint8)
    TIMESTAMP = np.empty(valid_line_count, dtype=np.float64)
    PCNT      = np.empty(valid_line_count, dtype=np.uint32)
    TCNT      = np.empty(valid_line_count, dtype=np.uint32)
    PWIDTH    = np.empty(valid_line_count, dtype=np.uint8)

    # Second pass: fill the preallocated arrays
    idx = 0
    with open(filepath, "r") as f:
        for line in f:
            if line.startswith("#Frame"):
                continue
            values = line.split("\t")
            # Convert values as needed and store them directly.
            BOARDID[idx]   = int(values[1])
            CHANNELID[idx] = int(values[2])
            TIMESTAMP[idx] = float(values[4])
            PCNT[idx]      = int(values[5])
            TCNT[idx]      = int(values[6])
            PWIDTH[idx]    = int(values[7])
            idx += 1

            # if idx >= valid_line_count:
            #     break

    data = {
        "BOARDID": BOARDID,
        "CHANNELID": CHANNELID,
        "TIMESTAMP": TIMESTAMP,
        "PCNT": PCNT,
        "TCNT": TCNT,
        "PWIDTH": PWIDTH
    }

    # mask = data["PWIDTH"] > 15
    # data = {key: data[key][mask] for key in data}

    data["PCNT"] = pcnt_correction(data["PCNT"])

    timeelapsed = np.amax(data["PCNT"]) - np.amin(data["PCNT"])

    if timeelapsed > 87000:
        print(f"Time elapsed: {timeelapsed}")

    data["TIMESTAMP"] = get_timestamp(data["PCNT"], data["TCNT"]).astype(np.uint64)

    sort = np.argsort(data["TIMESTAMP"])

    data = {key: data[key][sort] for key in data}


    data["UNIQUEID"] = get_unique_id(data["BOARDID"], data["CHANNELID"], unique_id_mapping)
    data["LAYERID"] = get_layer(data["BOARDID"], layer_mapping)
    data["EVENTID"] = label(data["TIMESTAMP"], event_threshold=event_threshold, hit_threshold=hit_threshold)

    events = event_grouping(data, event_selection)

    np.savez_compressed(f"{dir}/{file.split('.')[0]}.npz", events)

    del data


if __name__ == "__main__":
    filenames = ["/data9/MuographyPython/Data/Simulation/Geant4/modified_gaisser_9449/0"]
    unique_id_mapping = np.loadtxt("forward.txt", dtype=np.int64).reshape(26, 16)
    layer_mapping = np.loadtxt("layer.txt", dtype=np.int64)

    for filename in filenames:
        files = [f for f in os.listdir(filename) if f.endswith(".txt")]

        files = sorted(files)

        for file in tqdm(files):
            print(f"Processing {file}")
            parser(filename, file)

        # # Uncomment the following lines to use multiprocessing
        # with ProcessPoolExecutor() as executor:
        #     list(tqdm(executor.map(worker, [filename]*len(files), files), total=len(files)))

        print("Done")


from setuptools import setup, find_packages

setup(
    name='ncuhep',
    version='0.1.61',
    author='Kah Seng Phay',
    author_email='phay_ks@icloud.com',
    description='A Python project for NCU High Energy Physics.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/yourusername/ncuhep',
    packages=find_packages(),
    license='MIT',  # Add this line
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    python_requires='>=3.6',
    install_requires=[
        'opencv-python',
        'numpy',
        "scipy",
        "iminuit",
        "matplotlib",
        "tqdm",
        "numba",
    ],
)


from .coordinates import cart2slope, cart2slope_vec, slope2cart, slope2cart_vec, det2earth, det2earth_vec, earth2det, earth2det_vec, zenith_azimuth_deg, zenith_azimuth_deg_vec
from .flux import effective_area, solid_angle
from .hough_transformation import array2combo, intercept_, multiple_intercept_
from .tracking import track_reconstruction
from .monte_carlo import homogenous_generator, detection_simulation, run_homogenous_simulation, compute_basis

__all__ = [
    "cart2slope",
    "cart2slope_vec",
    "slope2cart",
    "slope2cart_vec",
    "det2earth",
    "det2earth_vec",
    "earth2det",
    "earth2det_vec",
    "zenith_azimuth_deg",
    "zenith_azimuth_deg_vec",
    "effective_area",
    "solid_angle",
    "array2combo",
    "intercept_",
    "multiple_intercept_",
    "track_reconstruction",
    "homogenous_generator",
    "detection_simulation",
    "run_homogenous_simulation",
    "compute_basis"
]